import {
  Runtime,
  getReactNativeVersion,
  DEFAULT_CONFIG_FILENAME,
} from '@haul-bundler/core';
import ora, { Ora } from 'ora';
import fs from 'fs';
import path from 'path';
import inquirer from 'inquirer';
import dedent from 'dedent';
import which from 'which';
import exec from 'execa';

const delay = (time: number) =>
  new Promise(resolve => setTimeout(resolve, time));

async function checkProject(progress: Ora, cwd: string, runtime: Runtime) {
  progress.start('Checking project files');
  await delay(1000);

  // Are we inside a React Native project?
  if (getReactNativeVersion(cwd)) {
    progress.succeed('Project looks good');
  } else {
    progress.fail(dedent`
    This doesn't seem to be a React Native project.

    Make sure you have a ${runtime.logger.enhanceWithModifier(
      'bold',
      'package.json'
    )} file with ${runtime.logger.enhanceWithModifier(
      'bold',
      'react-native'
    )} in dependencies, and you have installed these dependencies.

    To generate a React Native project, run ${runtime.logger.enhanceWithModifier(
      'bold',
      'react-native init <ProjectName>'
    )}. See ${runtime.logger.enhanceWithColor(
      'cyan',
      'https://facebook.github.io/react-native/docs/getting-started.html'
    )} for details.
  `);
    runtime.complete(1);
  }
}

async function createHaulProjectConfig(
  progress: Ora,
  cwd: string,
  runtime: Runtime,
  preset: string
) {
  // Does `haul.config.js` already exist?
  let overwrite = true;
  if (fs.existsSync(path.join(cwd, DEFAULT_CONFIG_FILENAME))) {
    overwrite = ((await inquirer.prompt([
      {
        type: 'confirm',
        name: 'overwrite',
        message: `There is already a '${DEFAULT_CONFIG_FILENAME}'. Overwrite it?`,
      },
    ])) as { overwrite: boolean }).overwrite;
  }

  if (!overwrite) {
    return;
  }

  progress.start('Generating config files');

  await delay(1000);

  const config = dedent`
    import { withPolyfills, makeConfig } from "${preset}";

    export default makeConfig({
      bundles: {
        index: {
          entry: withPolyfills('./index'),
        },
      },
    });
  `;

  fs.writeFileSync(path.join(cwd, DEFAULT_CONFIG_FILENAME), config);

  progress.succeed(
    `Generated ${runtime.logger.enhanceWithModifier(
      'bold',
      DEFAULT_CONFIG_FILENAME
    )}`
  );
}

async function modifyBabelConfig(
  progress: Ora,
  cwd: string,
  runtime: Runtime,
  babelPreset: string
) {
  progress.start('Updating Babel config');

  const defaultBabelConfigPaths = [
    path.join(cwd, 'babel.config.js'),
    path.join(cwd, '.babelrc.js'),
    path.join(cwd, '.babelrc'),
  ];

  let babelConfigPath = defaultBabelConfigPaths.find(filePath =>
    fs.existsSync(filePath)
  );

  if (!babelConfigPath) {
    const result = (await inquirer.prompt([
      {
        type: 'input',
        name: 'entry',
        message: 'Enter path to the Babel config file',
        validate: (pathToFile: string) =>
          fs.existsSync(pathToFile) ? true : `${pathToFile} is not a valid`,
      },
    ])) as { entry: string };

    babelConfigPath = path.resolve(result.entry);
  }

  const babelConfig = fs.readFileSync(babelConfigPath).toString();
  fs.writeFileSync(
    babelConfigPath,
    babelConfig.replace('metro-react-native-babel-preset', babelPreset)
  );

  progress.succeed(
    `Updated Babel config at ${runtime.logger.enhanceWithModifier(
      'bold',
      path.relative(cwd, babelConfigPath)
    )}`
  );
}

async function modifyXcodeProject(progress: Ora, cwd: string) {
  let xcodeProject;

  // Does `ios/*.xcodeproj` exist?
  const iosPath = path.join(cwd, 'ios');
  if (fs.existsSync(iosPath)) {
    xcodeProject = fs
      .readdirSync(iosPath)
      .find(file => file.includes('.xcodeproj'));
    xcodeProject = path.join(iosPath, xcodeProject || '');
  }

  // Otherwise, ask for path to a file
  if (!xcodeProject) {
    const result = (await inquirer.prompt([
      {
        type: 'input',
        name: 'entry',
        message: 'Enter path to the .xcodeproj file',
        validate: (pathToFile: string) =>
          fs.existsSync(pathToFile) && pathToFile.includes('.xcodeproj')
            ? true
            : `${pathToFile} is not a valid .xcodeproj`,
      },
    ])) as { entry: string };

    xcodeProject = path.resolve(result.entry);
  }

  progress.start('Adding haul to your Xcode build scripts');

  await delay(1000);

  let project = fs
    .readFileSync(path.join(xcodeProject, 'project.pbxproj'))
    .toString();

  const haulSignature = 'added by Haul';

  /* Make sure we check both project iOS and iOS-TV, that's the magic behind "2" constant */

  const PROJECTS_COUNT = 2;

  const countOccurrences = (search: string) => {
    return project.split(search).length - 1;
  };

  // Are we already integrated?
  if (countOccurrences(haulSignature) === PROJECTS_COUNT) {
    progress.info('Haul is already part of your build scripts');
    return;
  }

  const originalTask = 'shellScript = "export NODE_BINARY=node';

  if (countOccurrences(originalTask) !== PROJECTS_COUNT) {
    progress.warn(
      `Couldn't edit Xcode project. Haven't recognized 'Bundle React Native code and images' build phase.`
    );
    return;
  }

  const haulTask = `shellScript = "# ${haulSignature}\\nexport CLI_PATH=node_modules/@haul-bundler/cli/bin/haul.js\\nexport NODE_BINARY=node`;

  project = project.replace(new RegExp(originalTask, 'g'), haulTask);

  fs.writeFileSync(path.join(xcodeProject, 'project.pbxproj'), project);
  progress.succeed('Added haul to your Xcode build scripts');
}

async function modifyGradleBuild(progress: Ora, cwd: string) {
  let gradleBuildFile;
  // Does `android/app/build.gradle` exist?
  const androidPath = path.join(cwd, 'android/app');
  if (fs.existsSync(androidPath)) {
    gradleBuildFile = fs
      .readdirSync(androidPath)
      .find(file => file.includes('build.gradle'));
    gradleBuildFile = path.join(androidPath, gradleBuildFile || '');
  }
  // Otherwise, ask for path to a file
  if (!gradleBuildFile) {
    const result = (await inquirer.prompt([
      {
        type: 'input',
        name: 'entry',
        message: 'Enter path to the android/app/build.gradle file',
        validate: (pathToFile: string) =>
          fs.existsSync(pathToFile) && pathToFile.includes('build.gradle')
            ? true
            : `${pathToFile} is not a valid build.gradle`,
      },
    ])) as { entry: string };
    gradleBuildFile = path.resolve(result.entry);
  }
  progress.start('Adding haul to your build.gradle');
  await delay(1000);
  let project = fs.readFileSync(gradleBuildFile).toString();
  const cliString = '"node_modules/@haul-bundler/cli/bin/haul.js"';

  // Are we already integrated?
  if (project.includes(cliString)) {
    progress.info('Haul is already part of your build.gradle');
    return;
  }
  project = project.replace(
    /project\.ext\.react = \[\n(.+)\n\]/,
    dedent`
    project.ext.react = [
    $1,
        cliPath: ${cliString}
    ]
    `
  );
  fs.writeFileSync(gradleBuildFile, project);
  progress.succeed('Added haul to your build.gradle');
}

function getRunScript(scriptName: string) {
  return `yarn run ${scriptName}`;
}

async function addHaulScript(progress: Ora, cwd: string): Promise<string> {
  const packageJson = require(path.join(cwd, 'package.json'));
  const scripts = packageJson.scripts || {};
  const haulScript = Object.keys(scripts).find(
    name => scripts[name] === 'haul'
  );

  if (haulScript) {
    progress.info('Haul already exists in your package.json.');
    return haulScript;
  }

  let scriptName = 'start';
  if (
    scripts.start &&
    scripts.start !== 'node node_modules/react-native/local-cli/cli.js start'
  ) {
    const result = (await inquirer.prompt([
      {
        type: 'input',
        name: 'scriptName',
        message:
          'Enter the name of the script to add to package.json, e.g. `haul` for `yarn run haul`',
        default: 'haul',
      },
    ])) as { scriptName: string };

    scriptName = result.scriptName;
  }

  packageJson.scripts = Object.assign({}, scripts, {
    [scriptName]: 'haul start',
  });

  progress.start(`Adding \`${scriptName}\` script to your package.json`);

  await delay(1000);

  fs.writeFileSync(
    path.join(cwd, 'package.json'),
    JSON.stringify(packageJson, null, 2)
  );

  return scriptName;
}

async function installDependencies(
  progress: Ora,
  { babelPreset, haulPreset }: { babelPreset: string; haulPreset: string }
) {
  progress.info('Installing required devDependencies');
  const useYarn = await new Promise<boolean>(resolve => {
    which('yarn', (_, resolved: string | undefined) => {
      resolve(Boolean(resolved));
    });
  });
  const installArgs = (useYarn
    ? ['add', '-D']
    : ['install', '--save-dev']
  ).concat(babelPreset, haulPreset);
  await exec(useYarn ? 'yarn' : 'npm', installArgs, { stdio: 'inherit' });
}

export default function initCommand(runtime: Runtime) {
  return {
    command: 'init',
    describe: 'Generates necessary configuration files',
    async handler() {
      let exitCode = 0;
      try {
        const cwd = process.cwd();
        const rnVersion = getReactNativeVersion(cwd);

        if (!rnVersion) {
          runtime.logger.error(
            'Cannot find React Native. Are you in React Native project?'
          );
          runtime.complete(1);
        }

        const babelPreset = '@haul-bundler/babel-preset-react-native';
        const haulPreset = `@haul-bundler/preset-${rnVersion!.major}.${
          rnVersion!.minor
        }`;

        const progress = ora().start();

        await checkProject(progress, cwd, runtime);
        await createHaulProjectConfig(progress, cwd, runtime, haulPreset);
        await modifyBabelConfig(progress, cwd, runtime, babelPreset);
        await modifyXcodeProject(progress, cwd);
        await modifyGradleBuild(progress, cwd);
        const scriptName = await addHaulScript(progress, cwd);
        await installDependencies(progress, { babelPreset, haulPreset });
        progress.succeed(
          `You can now start Haul by running '${getRunScript(scriptName)}'`
        );
      } catch (error) {
        runtime.logger.error('Command failed with error:', error);
        exitCode = 1;
      } finally {
        runtime.complete(exitCode);
      }
    },
  };
}
